"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.DocumentManager = void 0;
const protocol_definitions_1 = require("@fluidframework/protocol-definitions");
const server_services_client_1 = require("@fluidframework/server-services-client");
const server_services_utils_1 = require("@fluidframework/server-services-utils");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
/**
 * Manager to fetch document from Alfred using the internal URL.
 */
class DocumentManager {
    constructor(internalAlfredUrl, tenantManager, documentStaticDataCache) {
        this.internalAlfredUrl = internalAlfredUrl;
        this.tenantManager = tenantManager;
        this.documentStaticDataCache = documentStaticDataCache;
        if (!this.documentStaticDataCache) {
            server_services_telemetry_1.Lumberjack.info("DocumentManager static data cache is undefined, cache will not be used.");
        }
    }
    async readDocument(tenantId, documentId) {
        // Retrieve the document
        const restWrapper = await this.getBasicRestWrapper(tenantId, documentId);
        const document = await restWrapper.get(`/documents/${tenantId}/${documentId}`);
        if (!document) {
            return undefined;
        }
        if (this.documentStaticDataCache) {
            // Cache the static properties of the document
            const staticProps = DocumentManager.getStaticPropsFromDoc(document);
            const staticPropsKey = DocumentManager.getDocumentStaticKey(documentId);
            await this.documentStaticDataCache.set(staticPropsKey, JSON.stringify(staticProps));
        }
        // Return the original document that was retrieved
        return document;
    }
    async readStaticProperties(tenantId, documentId) {
        // If the cache is undefined, fetch the document from the database
        if (!this.documentStaticDataCache) {
            server_services_telemetry_1.Lumberjack.verbose("Falling back to database after attempting to read cached static document data, because the DocumentManager cache is undefined.", (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId));
            const document = await this.readDocument(tenantId, documentId);
            return document;
        }
        // Retrieve cached static document props
        const staticPropsKey = DocumentManager.getDocumentStaticKey(documentId);
        const staticPropsStr = await this.documentStaticDataCache.get(staticPropsKey);
        // If there are no cached static document props, fetch the document from the database
        if (!staticPropsStr) {
            server_services_telemetry_1.Lumberjack.verbose("Falling back to database after attempting to read cached static document data.", (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId));
            const document = await this.readDocument(tenantId, documentId);
            return DocumentManager.getStaticPropsFromDoc(document);
        }
        // Return the static data, parsed into a JSON object
        const staticProps = JSON.parse(staticPropsStr);
        return staticProps;
    }
    async purgeStaticCache(documentId) {
        // If the cache is undefined, do nothing, because there are no cached static properties to purge
        if (!this.documentStaticDataCache) {
            server_services_telemetry_1.Lumberjack.error("Cannot purge document static properties cache, because the DocumentManager cache is undefined.");
            return;
        }
        const staticPropsKey = DocumentManager.getDocumentStaticKey(documentId);
        await this.documentStaticDataCache.delete(staticPropsKey);
    }
    async getBasicRestWrapper(tenantId, documentId) {
        const key = await this.tenantManager.getKey(tenantId);
        const getDefaultHeaders = () => {
            const jwtToken = (0, server_services_utils_1.generateToken)(tenantId, documentId, key, [protocol_definitions_1.ScopeType.DocRead]);
            return {
                Authorization: `Basic ${jwtToken}`,
            };
        };
        const restWrapper = new server_services_client_1.BasicRestWrapper(this.internalAlfredUrl, undefined /* defaultQueryString */, undefined /* maxBodyLength */, undefined /* maxContentLength */, getDefaultHeaders(), undefined /* Axios */, undefined /* refreshDefaultQueryString */, getDefaultHeaders /* refreshDefaultHeaders */, server_services_utils_1.getCorrelationId /* getCorrelationId */);
        return restWrapper;
    }
    /**
     * Creates a cache key to retreive static data from a document
     *
     * @param documentId - ID of the document to create an access key for
     * @returns - A cache key to access static data for [documentId]
     */
    static getDocumentStaticKey(documentId) {
        return `staticData:${documentId}`;
    }
    /**
     * Extracts the static properties from an IDocument
     *
     * @param document - Document to get properties from
     * @returns - The static properties of [document]
     */
    static getStaticPropsFromDoc(document) {
        return {
            version: document.version,
            createTime: document.createTime,
            documentId: document.documentId,
            tenantId: document.tenantId,
            storageName: document.storageName,
            isEphemeralContainer: document.isEphemeralContainer,
        };
    }
}
exports.DocumentManager = DocumentManager;
//# sourceMappingURL=documentManager.js.map